#!/usr/bin/env python3
"""
Proxy Server
HTTP/HTTPS proxy server with traffic inspection capabilities.
Parameters:
    - host: Host to bind the proxy server
    - port: Port to listen on
    - cert: Path to SSL certificate for HTTPS
    - key: Path to SSL private key
Requires: ssl package for HTTPS support
"""
import socket
import select
import threading
import argparse
import logging
import ssl
from http.server import HTTPServer, BaseHTTPRequestHandler
from urllib.parse import urlparse
import queue
import time

class ProxyHandler(BaseHTTPRequestHandler):
    timeout = 5
    lock = threading.Lock()

    def __init__(self, *args, **kwargs):
        self.tls_ctx = ssl.create_default_context()
        self.tls_ctx.check_hostname = False
        self.tls_ctx.verify_mode = ssl.CERT_NONE
        super().__init__(*args, **kwargs)

    def do_CONNECT(self):
        address = self.path.split(':')
        address[1] = int(address[1]) or 443
        
        try:
            s = socket.create_connection(address, timeout=self.timeout)
            self.send_response(200, 'Connection Established')
            self.end_headers()

            if address[1] == 443:
                self.connection = self.tls_ctx.wrap_socket(self.connection,
                                                         server_side=True,
                                                         certfile=self.server.cert_file,
                                                         keyfile=self.server.key_file)

            self.handle_tcp(s, address)
        except Exception as e:
            self.send_error(500, str(e))
            return
        finally:
            s.close()

    def do_GET(self):
        url = urlparse(self.path)
        address = (url.hostname, url.port or 80)
        
        try:
            s = socket.create_connection(address, timeout=self.timeout)
            
            if url.scheme == 'https':
                s = self.tls_ctx.wrap_socket(s, server_hostname=url.hostname)
            
            s.send(f"{self.command} {url.path} {self.protocol_version}\r\n".encode())
            self.headers['Connection'] = 'close'
            for key, value in self.headers.items():
                s.send(f"{key}: {value}\r\n".encode())
            s.send(b"\r\n")
            
            self.handle_tcp(s, address)
        except Exception as e:
            self.send_error(500, str(e))
            return
        finally:
            s.close()

    def handle_tcp(self, sock, address):
        socks = [self.connection, sock]
        count = 0
        while True:
            count += 1
            (recv, _, error) = select.select(socks, [], socks, self.timeout)
            if error:
                break
            if not recv:
                break
            for in_ in recv:
                other = self.connection if in_ is sock else sock
                try:
                    data = in_.recv(8192)
                except ConnectionResetError:
                    return
                if not data:
                    return
                other.send(data)

class ThreadedHTTPServer(HTTPServer):
    def __init__(self, server_address, RequestHandlerClass, cert_file=None, key_file=None):
        super().__init__(server_address, RequestHandlerClass)
        self.cert_file = cert_file
        self.key_file = key_file
        self.daemon_threads = True

class ProxyServer:
    def __init__(self, host='localhost', port=8080, cert_file=None, key_file=None):
        self.host = host
        self.port = port
        self.cert_file = cert_file
        self.key_file = key_file
        self.setup_logging()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def start(self):
        try:
            server_address = (self.host, self.port)
            self.httpd = ThreadedHTTPServer(server_address, ProxyHandler,
                                          self.cert_file, self.key_file)
            
            if self.cert_file and self.key_file:
                self.httpd.socket = ssl.wrap_socket(self.httpd.socket,
                                                  certfile=self.cert_file,
                                                  keyfile=self.key_file,
                                                  server_side=True)
            
            self.logger.info(f"Starting proxy server on {self.host}:{self.port}")
            if self.cert_file and self.key_file:
                self.logger.info("HTTPS support enabled")
            
            self.httpd.serve_forever()
            
        except KeyboardInterrupt:
            self.logger.info("Stopping proxy server...")
            self.httpd.server_close()
        except Exception as e:
            self.logger.error(f"Error starting proxy server: {str(e)}")

class RequestLogger:
    def __init__(self, output_file=None):
        self.output_file = output_file
        self.log_queue = queue.Queue()
        self.setup_logging()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def log_request(self, request):
        timestamp = time.strftime('%Y-%m-%d %H:%M:%S')
        log_entry = {
            'timestamp': timestamp,
            'method': request.command,
            'path': request.path,
            'headers': dict(request.headers),
            'client_address': request.client_address[0]
        }
        self.log_queue.put(log_entry)

    def start_logging(self):
        while True:
            try:
                log_entry = self.log_queue.get()
                if self.output_file:
                    with open(self.output_file, 'a') as f:
                        f.write(f"{log_entry}\n")
                self.logger.info(f"Request: {log_entry['method']} {log_entry['path']}")
            except Exception as e:
                self.logger.error(f"Error logging request: {str(e)}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="HTTP/HTTPS Proxy Server")
    parser.add_argument("--host", default="localhost",
                      help="Host to bind the proxy server")
    parser.add_argument("--port", type=int, default=8080,
                      help="Port to listen on")
    parser.add_argument("--cert", help="Path to SSL certificate for HTTPS")
    parser.add_argument("--key", help="Path to SSL private key")
    args = parser.parse_args()

    if bool(args.cert) != bool(args.key):
        parser.error("Both --cert and --key must be provided for HTTPS support")

    proxy = ProxyServer(args.host, args.port, args.cert, args.key)
    proxy.start()
